"""Logging and alerts."""
import sys
from enum import Enum
from pathlib import Path
from textwrap import wrap

import rich.repr
import typer
from loguru import logger

from jdfile.utils.console import console


class LogLevel(Enum):
    """Enum for log levels."""

    TRACE = 5
    DEBUG = 10
    INFO = 20
    SUCCESS = 25
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    EXCEPTION = 60


class VerboseLevel(Enum):
    """Enum for verbose levels."""

    WARN = 0
    INFO = 1
    DEBUG = 2
    TRACE = 3


def dryrun(msg: str) -> None:
    """Print a message if the dry run flag is set.

    Args:
        msg: Message to print
    """
    console.print(f"[cyan]DRYRUN   | {msg}[/cyan]")


def success(msg: str) -> None:
    """Print a success message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[green]SUCCESS  | {msg}[/green]")


def warning(msg: str) -> None:
    """Print a warning message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[yellow]WARNING  | {msg}[/yellow]")


def error(msg: str) -> None:
    """Print an error message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[red]ERROR    | {msg}[/red]")


def notice(msg: str) -> None:
    """Print a notice message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[bold]NOTICE   | {msg}[/bold]")


def info(msg: str) -> None:
    """Print a notice message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"INFO     | {msg}")


def usage(msg: str, width: int = None) -> None:
    """Print a usage message without using logging.

    Args:
        msg: Message to print
        width (optional): Width of the message
    """
    if width is None:
        width = console.width - 15

    for _n, line in enumerate(wrap(msg, width=width)):
        if _n == 0:
            console.print(f"[dim]USAGE    | {line}")
        else:
            console.print(f"[dim]         | {line}")


def debug(msg: str) -> None:
    """Print a debug message without using logging.

    Args:
        msg: Message to print
    """
    console.print(f"[blue]DEBUG    | {msg}[/blue]")


def dim(msg: str) -> None:
    """Print a message in dimmed color.

    Args:
        msg: Message to print
    """
    console.print(f"[dim]{msg}[/dim]")


def _log_formatter(record: dict) -> str:
    """Create custom log formatter based on the log level.  This effects the logs sent to stdout/stderr but not the log file."""
    if (
        record["level"].name == "INFO"
        or record["level"].name == "SUCCESS"
        or record["level"].name == "WARNING"
    ):
        return "<level><normal>{level: <8} | {message}</normal></level>\n{exception}"

    if record["level"].name == "TRACE" or record["level"].name == "DEBUG":
        return "<level><normal>{level: <8} | {message}</normal></level> <fg #c5c5c5>({name}:{function}:{line})</fg #c5c5c5>\n{exception}"

    return "<level>{level: <8} | {message}</level> <fg #c5c5c5>({name}:{function}:{line})</fg #c5c5c5>\n{exception}"


@rich.repr.auto
class LoggerManager:
    """Instantiate the loguru logging system with the following levels.

        - TRACE: Usage: log.trace("")
        - DEBUG: Usage: log.debug("")
        - INFO: Usage: log.info("")
        - WARNING: Usage: log.warning("")
        - ERROR: Usage: log.error("")
        - CRITICAL: Usage: log.critical("")
        - EXCEPTION: Usage: log.exception("")

    Attributes:
        log_file (Path): Path to the log file.
        verbosity (int): Verbosity level.
        log_to_file (bool): Whether to log to a file.
        log_level (int): Default log level (verbosity overrides this)

    Examples:
        Instantiate the logger:

            logging = _utils.alerts.LoggerManager(
                verbosity,
                log_to_file,
                log_file,
                log_level)
    """

    def __init__(
        self,
        log_file: Path = Path("/logs"),
        verbosity: int = 0,
        log_to_file: bool = False,
        log_level: int = 30,
    ) -> None:
        self.verbosity = verbosity
        self.log_to_file = log_to_file
        self.log_file = log_file
        self.log_level = log_level

        if self.log_file == Path("/logs") and self.log_to_file:  # pragma: no cover
            print("No log file specified")
            raise typer.Exit(1)

        if self.verbosity >= VerboseLevel.TRACE.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="TRACE",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 5
        elif self.verbosity == VerboseLevel.DEBUG.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="DEBUG",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 10
        elif self.verbosity == VerboseLevel.INFO.value:
            logger.remove()
            logger.add(
                sys.stderr,
                level="INFO",
                format=_log_formatter,  # type: ignore[arg-type]
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 20
        else:
            logger.remove()
            logger.add(
                sys.stderr,
                format=_log_formatter,  # type: ignore[arg-type]
                level="SUCCESS",
                backtrace=False,
                diagnose=True,
            )
            self.log_level = 25

        if self.log_to_file is True:
            logger.add(
                self.log_file,
                rotation="5 MB",
                level=self.log_level,
                backtrace=False,
                diagnose=True,
                delay=True,
            )
            logger.debug(f"Logging to file: {self.log_file}")

        logger.debug("Logging instantiated")

    def is_trace(self, msg: str | None = None) -> bool:
        """Check if the current log level is TRACE.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is TRACE or lower, False otherwise.
        """
        if self.log_level <= LogLevel.TRACE.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_debug(self, msg: str | None = None) -> bool:
        """Check if the current log level is DEBUG.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is DEBUG or lower, False otherwise.
        """
        if self.log_level <= LogLevel.DEBUG.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_info(self, msg: str | None = None) -> bool:
        """Check if the current log level is INFO.

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is INFO or lower, False otherwise.
        """
        if self.log_level <= LogLevel.INFO.value:
            if msg:
                console.print(msg)
            return True
        return False

    def is_default(self, msg: str | None = None) -> bool:
        """Check if the current log level is default level (SUCCESS or WARNING).

        Args:
            msg (optional): Message to print. Defaults to None.

        Returns:
            bool: True if the current log level is default or lower, False otherwise.
        """
        if self.log_level <= LogLevel.WARNING.value:
            if msg:
                console.print(msg)
            return True
        return False  # pragma: no cover



"""Utilities for jdfile."""


import re
from pathlib import Path

from rich import box
from rich.table import Table

from jdfile._config.config import Config
from jdfile.models.file import File
from jdfile.models.project import Project
from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.console import console


def build_file_list(
    files: list[Path], config: Config, project: Project = None
) -> tuple[list[File], list[File]]:
    """Build a list of files to process. If a directory is specified, all files in the directory will be processed.

    Args:
        config (Config): Configuration object.
        files (list[Path]): List of files or directories to process.
        project (Project, optional): Project object. Defaults to None.

    Returns:
            list[Path]: List of files to process.
    """
    config._ignored_files.extend([".DS_Store", ".jdfile", ".stignore"])
    if len(config.ignored_regex) > 0:
        combined_regex = "(" + ")|(".join(config.ignored_regex) + ")"
    else:
        combined_regex = "^$"  # match nothing

    directories = [f for f in files if f.is_dir() and f.exists()]
    for _dir in directories:
        files.remove(_dir)
        for f in _dir.rglob("*"):
            depth_of_file = len(f.relative_to(_dir).parts)
            if depth_of_file <= config.depth and f.is_file():
                files.append(f)

    initial_files_length = len(files)

    with console.status(
        "Processing Files...  [dim](Can take a while for large directory trees)[/]",
        spinner="bouncingBall",
    ):
        all_files = [
            File(path=f, config=config, project=project)
            for f in files
            if f.is_file()
            and f.name not in config.ignored_files
            and not re.search(combined_regex, f.name)
        ]

    for file in [x for x in all_files if len(x.organize_possible_folders) > 0]:
        file.select_new_parent()

    if config.ignore_dotfiles:
        all_files = [f for f in all_files if not f.is_dotfile]

    processed_files = [f for f in all_files if f.organize_skip is False]
    skip_organize_files = [f for f in all_files if f.organize_skip is True]

    log.info(f"Processed {len(processed_files)} of {initial_files_length} files")
    return sorted(processed_files, key=lambda x: x.name), sorted(
        skip_organize_files, key=lambda x: x.name
    )


def table_confirmation(files: list[File], project: Project, total_files: int) -> None:
    """Display a confirmation table to the user.

    Args:
        files (list[File]): List of files to process.
        project (Project): Project object.
        total_files (int): Total number of files to process.
    """
    alerts.notice("Confirm the changes below")
    if project and project.exists and len([x for x in files if x.parent != x.new_parent]) > 0:
        organized_files = True
    else:
        organized_files = False

    table = Table(
        "#",
        "Original Name",
        "New Name",
        "New Path" if organized_files else "",
        "Diff",
        box=box.SIMPLE,
        show_header=True,
        header_style="bold",
        min_width=40,
        title=f"Pending changes for {len(files)} of {total_files} files",
    )
    for _n, file in enumerate(files, start=1):
        table.add_row(
            str(_n),
            file.stem + "".join(file.suffixes),
            file.new_stem + "".join(file.new_suffixes)
            if file.has_changes()
            else "[green]No Changes[/green]",
            str(file.new_parent.relative_to(project.path))
            if organized_files and file.parent != file.new_parent
            else "",
            file.print_diff() if file.has_changes() else "",
        )
    console.print(table)


def table_skipped_files(files: list[File]) -> None:
    """Display a table of files that could not be organized.

    Args:
        files (list[File]): List of files to process.
    """
    table = Table(
        "#",
        "Filename",
        box=None,
        show_header=False,
        title_style="yellow",
        title_justify="left",
        min_width=64,
    )
    alerts.warning("These files matched no folders and were skipped")
    for file in files:
        if len(file.path.parents) > 2:  # noqa: PLR2004
            parents = "         [dim]…/[/]" + str(
                file.path.parent.relative_to(file.path.parents[2])
            )
        else:
            parents = str(file.path.parent)

        table.add_row(f"{parents}/[bold]{file.name}[/]")

    console.print(table)
    console.print("\n")



"""Instantiate Configuration class and set default values."""

import functools
import re
import shutil
from pathlib import Path
from typing import Any

import rich.repr
import typer

from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.enums import InsertLocation, Separator, TransformCase

try:
    import tomllib
except ModuleNotFoundError:  # pragma: no cover
    import tomli as tomllib  # type: ignore [no-redef]

PATH_CONFIG_DEFAULT = Path(__file__).parent / "default_config.toml"


@rich.repr.auto
class Config:
    """Representation of a configuration file."""

    def __init__(
        self,
        config_path: Path = None,
        context: dict[str, Any] = {},
    ) -> None:
        """Initialize configuration file."""
        self.config_path = config_path.expanduser().resolve() if config_path else None
        self.context = context

        if (
            "project_name" in self.context
            and self.context["project_name"] is not None
            and config_path is not None
        ):
            if not self.config_path.exists():
                self._create_config()
            self.config_file_dict = self._load_config_file()
            self.project_config = self._find_project_config()
            self.project_path = Path(self.project_config["path"]).expanduser().resolve()
        else:
            self.project_name: str | None = None
            self.project_path = None
            self.config_file_dict = {}
            self.project_config = {}

        self._clean = (
            self.context["clean"]
            if "clean" in self.context and self.context["clean"] is not None
            else False
        )
        self._depth = (
            self.context["depth"]
            if "depth" in self.context and self.context["depth"] is not None
            else 1
        )
        self._ignored_files = (
            self.project_config["ignored_files"] if "ignored_files" in self.project_config else []
        )
        self._ignored_regex = (
            self.project_config["ignored_regex"] if "ignored_regex" in self.project_config else []
        )
        self._organize = (
            self.context["organize"]
            if "organize" in self.context and self.context["organize"] is not None
            else True
        )

        self._cli_terms = self._get_cli_terms()
        self._date_format = self._get_date_format()
        self._strip_stopwords = self._get_strip_stopwords()
        self._ignore_dotfiles = self._get_ignore_dotfiles()
        self._overwrite_existing = self._get_overwrite_existing()
        # TODO: Add insert location to cli/config options
        self.insert_location = InsertLocation.BEFORE

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Return the representation of the configuration file."""
        yield "config_path", self.config_path
        yield "clean", self.clean
        yield "cli_terms", self.cli_terms
        yield "date_format", self.date_format
        yield "dry_run", self.dry_run
        yield "force", self.force
        yield "ignore_dotfiles", self.ignore_dotfiles
        yield "ignored_files", self.ignored_files
        yield "ignored_regex", self.ignored_regex
        yield "insert_location", self.insert_location
        yield "match_case", self.match_case
        yield "organize", self.organize
        yield "project_name", self.project_name
        yield "project_path", self.project_path
        yield "separator", self.separator
        yield "split_words", self.split_words
        yield "strip_stopwords", self.strip_stopwords
        yield "stopwords", self.stopwords
        yield "transform_case", self.transform_case
        yield "use_nltk", self.use_nltk

    def _create_config(self) -> None:
        """Create a configuration file from the default when it does not exist.

        Returns:
            None: Exit the program after creating the configuration file.

        Raises:
            typer.Exit: Exit the program after creating the configuration file.
        """
        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(PATH_CONFIG_DEFAULT, self.config_path)
        alerts.success(f"Created default configuration file at {self.config_path}")
        alerts.notice(f"Please edit {self.config_path} before continuing")
        raise typer.Exit(code=1)

    def _get_cli_terms(self) -> list[str]:
        """Get the terms from the command line.

        Returns:
            list[str]: Terms from the command line.
        """
        if "cli_terms" in self.context and self.context["cli_terms"] is not None:
            return self.context["cli_terms"]

        return []

    def _get_date_format(self) -> str:
        """Get the date format for the project.

        Returns:
            str: Date format for the project.

        Raises:
            typer.Abort: If the date format is invalid.
        """
        if "date_format" in self.context and self.context["date_format"] is not None:
            date_format = self.context["date_format"]
        elif "date_format" in self.project_config:
            if self.project_config["date_format"].lower() == "none":
                return None
            date_format = self.project_config["date_format"]
        else:
            return None

        valid_format_codes = [
            "%a",
            "%A",
            "%w",
            "%d",
            "%b",
            "%B",
            "%m",
            "%y",
            "%Y",
            "%H",
            "%I",
            "%p",
            "%M",
            "%S",
            "%f",
            "%z",
            "%Z",
            "%j",
            "%U",
            "%W",
            "%c",
            "%x",
            "%X",
            "%%",
            "%G",
            "%u",
            "%V",
        ]
        values = re.findall(r"%\w", date_format)
        for value in values:
            if value not in valid_format_codes:
                alerts.error(f"Invalid date format: {date_format}")
                raise typer.Abort()

        return date_format

    def _get_overwrite_existing(self) -> bool:
        """Get the overwrite_existing option for the project."""
        if "overwrite_existing" in self.context and self.context["overwrite_existing"] is not None:
            return self.context["overwrite_existing"]

        if "overwrite_existing" in self.project_config:
            return self.project_config["overwrite_existing"]

        return False

    def _get_strip_stopwords(self) -> bool:
        """Strip stop words from a string.

        Returns:
            bool: Strip stopwords option for the project.
        """
        if "strip_stopwords" in self.context and isinstance(self.context["strip_stopwords"], bool):
            return self.context["strip_stopwords"]

        if "strip_stopwords" in self.project_config:
            return self.project_config["strip_stopwords"]

        return False

    def _get_ignore_dotfiles(self) -> bool:
        """Get the ignore_dotfiles option for the project.

        Returns:
            bool: ignore_dotfiles option for the project.
        """
        if "ignore_dotfiles" in self.context and self.context["ignore_dotfiles"] is not None:
            return self.context["ignore_dotfiles"]

        if "ignore_dotfiles" in self.project_config:
            return self.project_config["ignore_dotfiles"]

        return True

    def _find_project_config(self) -> dict[str, Any]:
        """Find a project configuration in the config file matching the project name.

        Returns:
            dict[str, Any]: Configuration for this run.

        Raises:
            typer.Abort: If the project does not exist in the config file.
        """
        for _project in self.config_file_dict:
            if self.context["project_name"] == _project:
                self.project_name = _project
                return self.config_file_dict[_project]

        log.error(f"Project does not exist in config file: {self.context['project_name']}")
        raise typer.Abort()

    def _load_config_file(self) -> dict[str, Any]:
        """Load the configuration file.

        Returns:
            dict[str, Any]: Configuration file as a dictionary.

        Raises:
            typer.Exit: If the configuration file is empty or malformed.
        """
        log.debug(f"Loading configuration from {self.config_path}")
        with self.config_path.open("rb") as f:
            try:
                config_file_dict = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                log.exception(f"Could not parse '{self.config_path}'")
                raise typer.Exit(code=1) from e

        if config_file_dict == {}:
            log.error(f"Configuration file '{self.config_path}' is empty or malformed")
            raise typer.Exit(code=1)

        for key, value in config_file_dict.items():
            if not isinstance(value, dict) or "path" not in config_file_dict[key]:
                log.error(f"Configuration file '{self.config_path}' is malformed")
                raise typer.Exit(code=1)

        return config_file_dict

    @functools.cached_property
    def transform_case(self) -> TransformCase:
        """Get the case transformation for the project.

        Returns:
            TransformCase: Case transformation for the project.

        Raises:
            typer.Abort: If the case transformation is invalid.
        """
        if "transform_case" in self.context and self.context["transform_case"] is not None:
            transform_case = self.context["transform_case"]
        elif "transform_case" in self.project_config:
            transform_case = self.project_config["transform_case"]
        else:
            return TransformCase.IGNORE

        if isinstance(transform_case, str):
            try:
                transform_case = TransformCase[transform_case.upper()]
            except KeyError as e:
                alerts.error(f"Invalid case: {transform_case}")
                raise typer.Abort() from e

        return transform_case

    @functools.cached_property
    def separator(self) -> Separator:
        """Get the separator transformation for the project.

        Returns:
            Separator: Separator transformation for the project.

        Raises:
            typer.Abort: If the separator transformation is invalid.
        """
        if "separator" in self.context and self.context["separator"] is not None:
            separator = self.context["separator"]
        elif "separator" in self.project_config:
            separator = self.project_config["separator"]
        else:
            return Separator.IGNORE

        if isinstance(separator, str):
            try:
                separator = Separator[separator.upper()]
            except KeyError as e:
                alerts.error(f"Invalid separator: {separator}")
                raise typer.Abort() from e

        return separator

    @property
    def clean(self) -> bool:
        """Get the clean option for the project.

        Returns:
            bool: Clean option for the project.
        """
        return self._clean

    @clean.setter
    def clean(self, value: bool) -> None:
        self._clean = value

    @property
    def date_format(self) -> str:
        """Get the date format for the project.

        Returns:
            str: Date format for the project.
        """
        return self._date_format

    @date_format.setter
    def date_format(self, value: str) -> None:
        self._date_format = value

    @property
    def dry_run(self) -> bool:
        """Get the dry run option for the project.

        Returns:
            bool: Dry run option for the project.
        """
        if "dry_run" in self.context and self.context["dry_run"] is not None:
            return self.context["dry_run"]

        return False

    @property
    def force(self) -> bool:
        """Get the force option for the project.

        Returns:
            bool: Force option for the project.
        """
        if "force" in self.context and self.context["force"] is not None:
            return self.context["force"]

        return False

    @property
    def ignore_dotfiles(self) -> bool:
        """Get the ignore_dotfiles option for the project.

        Returns:
            bool: ignore_dotfiles option for the project.
        """
        return self._ignore_dotfiles

    @ignore_dotfiles.setter
    def ignore_dotfiles(self, value: bool) -> None:
        self._ignore_dotfiles = value

    @property
    def ignored_files(self) -> list[str]:
        """Get the ignored files for the project."""
        return self._ignored_files

    @ignored_files.setter
    def ignored_files(self, value: list[str]) -> None:
        self._ignored_files = value

    @property
    def ignored_regex(self) -> list[str]:
        """Get the ignored regex patterns for the project."""
        return self._ignored_regex

    @ignored_regex.setter
    def ignored_regex(self, value: list[str]) -> None:
        self._ignored_regex = value

    @property
    def split_words(self) -> bool:
        """Split words in a string.

        Returns:
            bool: Split words option for the project.
        """
        if "split_words" in self.context and isinstance(self.context["split_words"], bool):
            return self.context["split_words"]

        if "split_words" in self.project_config:
            return self.project_config["split_words"]

        return False

    @property
    def cli_terms(self) -> list[str]:
        """Get the terms from the CLI.

        Returns:
            List[str]: Terms from the CLI.
        """
        return self._cli_terms

    @cli_terms.setter
    def cli_terms(self, value: list[str]) -> None:
        self._cli_terms = value

    @property
    def depth(self) -> int:
        """Set the depth to search for files.

        Returns:
            int: Depth to search for files.
        """
        return self._depth

    @depth.setter
    def depth(self, value: int) -> None:
        self._depth = value

    @property
    def strip_stopwords(self) -> bool:
        return self._strip_stopwords

    @strip_stopwords.setter
    def strip_stopwords(self, value: bool) -> None:
        self._strip_stopwords = value

    @property
    def match_case(self) -> list[str]:
        """Get the match case for the project.

        Returns:
            List[str]: Match case for the project.
        """
        if "match_case" in self.project_config:
            return self.project_config["match_case"]

        return []

    @property
    def organize(self) -> bool:
        """Get the organize option for the project.

        Returns:
            bool: Organize option for the project.
        """
        return self._organize

    @organize.setter
    def organize(self, value: bool) -> None:
        self._organize = value

    @property
    def overwrite_existing(self) -> bool:
        """Get the overwrite existing option for the project.

        Returns:
            bool: Overwrite existing option for the project.
        """
        return self._overwrite_existing

    @overwrite_existing.setter
    def overwrite_existing(self, value: bool) -> None:
        self._overwrite_existing = value

    @property
    def stopwords(self) -> list[str]:
        """Get the stopwords for the project.

        Returns:
            List[str]: Stopwords for the project.
        """
        if "stopwords" in self.project_config:
            return self.project_config["stopwords"]

        return []

    @property
    def use_nltk(self) -> bool:
        """Get the use nltk option for the project.

        Returns:
            bool: Use nltk option for the project.
        """
        if "use_nltk" in self.context and self.context["use_nltk"] is not None:
            return self.context["use_nltk"]

        if "use_synonyms" in self.project_config:
            return self.project_config["use_synonyms"]

        return False



"""Gather user input from the command line."""

from pathlib import Path

import questionary
import typer

from jdfile.utils import alerts

questionary.prompts.select.DEFAULT_STYLE = questionary.Style([("qmark", "")])
STYLE = questionary.Style(
    [
        ("qmark", "bold"),
        ("question", "bold"),
        ("separator", "fg:#808080"),
        ("answer", "fg:#FF9D00 bold"),
        ("instruction", "fg:#808080"),
        ("highlighted", "bold underline"),
        ("text", ""),
        ("pointer", "bold"),
    ]
)


def select_folder(
    possible_folders: dict, project_path: Path, filename: str
) -> str:  # pragma: no cover
    """Select a folder from a list of choices.

    Args:
        filename (str): Name of the file.
        possible_folders (dict): Dictionary of possible folders.
        project_path (Path): Path to the root of the project.

    Returns:
        str: Path to the selected folder or "skip"
    """
    choices: list[dict[str, str] | questionary.Separator] = [questionary.Separator()]
    max_length = len(max(possible_folders, key=len))

    for _k, _v in possible_folders.items():
        matching_terms = ", ".join(set(_v[1]))
        folder_path = str(_v[0].path.relative_to(project_path))

        choices.append(
            {
                "name": f"{folder_path:{max_length}} [matching: {matching_terms}]",
                "value": _v[0].path,
            }
        )

    choices.extend(
        [
            questionary.Separator(),
            {"name": "Skip this file", "value": "skip"},
            {"name": "Abort", "value": "abort"},
        ]
    )

    alerts.notice(f"Found {len(possible_folders)} possible folders for '[cyan bold]{filename}[/]'")
    result = questionary.select(
        "Select a folder", choices=choices, style=STYLE, qmark="INPUT    |"
    ).ask()

    if result is None or result == "abort":
        raise typer.Abort()

    return result



"""Date class for jdfile."""
import re
from dataclasses import dataclass
from datetime import date, datetime, timedelta

from jdfile.utils.alerts import logger as log


@dataclass
class DatePattern:
    """Regex patterns to find dates in filename strings."""

    string: str
    pattern_day_flexible = r"0?[1-9]|[12][0-9]|3[01]"
    pattern_day_inflexible = r"0[1-9]|[12][0-9]|3[01]"
    pattern_month = r"0[1-9]|1[012]"
    pattern_months = r"january|jan?|february|feb?|march|mar?|april|apr?|may|june?|july?|august|aug?|september|sep?t?|october|oct?|november|nov?|december|dec?"
    pattern_separator = r"[-\./_, :]*?"
    pattern_year = r"20[0-2][0-9]"

    def _month_to_number(self, month: str) -> str:  # noqa: C901, PLR0911
        """Convert a month name to a number.

        Args:
            month: (str) The month to convert.

        Returns:
            (str) The month number.
        """
        if re.match(r"^ja.*", month, re.I):
            return "01"
        if re.match(r"^fe.*", month, re.I):
            return "02"
        if re.match(r"^mar.*", month, re.I):
            return "03"
        if re.match(r"^ap.*", month, re.I):
            return "04"
        if re.match(r"^may.*", month, re.I):
            return "05"
        if re.match(r"^jun.*", month, re.I):
            return "06"
        if re.match(r"^jul.*", month, re.I):
            return "07"
        if re.match(r"^au.*", month, re.I):
            return "08"
        if re.match(r"^se.*", month, re.I):
            return "09"
        if re.match(r"^oc.*", month, re.I):
            return "10"
        if re.match(r"^no.*", month, re.I):
            return "11"
        if re.match(r"^de.*", month, re.I):
            return "12"

        return ""

    def yyyy_mm_dd(self) -> tuple[date, str] | None:
        """Search for a date in the format yyyy-mm-dd.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            # (?:.*[^0-9]|^)        # Start of string
            (?P<found>
                (?P<year>{self.pattern_year})
                {self.pattern_separator}
                (?P<month>{self.pattern_month})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_inflexible})
            )
            # (?:[^0-9].*|$)        # End of string from end of date
            """,
            re.X,
        )
        match = pattern.search(self.string)
        if match:
            try:
                return (
                    date(
                        int(match.group("year")), int(match.group("month")), int(match.group("day"))
                    ),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def yyyy_dd_mm(self) -> tuple[date, str] | None:
        """Search for a date in the format yyyy-dd-mm.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            # (?:.*[^0-9]|^)
            (?P<found>
                (?P<year>{self.pattern_year})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_inflexible})
                {self.pattern_separator}
                (?P<month>{self.pattern_month})
            )
            # (?:[^0-9].*|$)
            """,
            re.X,
        )
        match = pattern.search(self.string)
        if match:
            try:
                return (
                    date(
                        int(match.group("year")), int(match.group("month")), int(match.group("day"))
                    ),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def month_dd_yyyy(self) -> tuple[date, str] | None:
        """Search for a date in the format month dd, yyyy.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<month>{self.pattern_months})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_flexible})(?:nd|rd|th|st)?
                {self.pattern_separator}
                (?P<year>{self.pattern_year})
            )
            ([^0-9].*|$) # End of string from end of date)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            month = int(self._month_to_number(match.group("month")))
            try:
                return (
                    date(int(match.group("year")), month, int(match.group("day"))),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def dd_month_yyyy(self) -> tuple[date, str] | None:
        """Search for a date in the format dd month yyyy.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?:.*[^0-9]|^) # text before date
            (?P<found>
                (?P<day>{self.pattern_day_flexible})(?:nd|rd|th|st)?
                {self.pattern_separator}
                (?P<month>{self.pattern_months})
                {self.pattern_separator}
                (?P<year>{self.pattern_year})
            )
            (?:[^0-9].*|$) # text after date (7)
        """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            month = int(self._month_to_number(match.group("month")))
            try:
                return (
                    date(int(match.group("year")), month, int(match.group("day"))),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def month_dd(self) -> tuple[date, str] | None:
        """Search for a date in the format month dd.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<month>{self.pattern_months})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_flexible})(?:nd|rd|th|st)?
            )
            ([^0-9].*|$) # End of string from end of date)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            month = int(self._month_to_number(match.group("month")))
            year = date.today().year
            try:
                return (
                    date(year, month, int(match.group("day"))),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def month_yyyy(self) -> tuple[date, str] | None:
        """Search for a date in the format month yyyy.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<month>{self.pattern_months})
                {self.pattern_separator}
                (?P<year>{self.pattern_year})
            )
            ([^0-9].*|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            month = int(self._month_to_number(match.group("month")))
            try:
                return (
                    date(int(match.group("year")), month, 1),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def yyyy_month(self) -> tuple[date, str] | None:
        """Search for a date in the format yyyy month.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<year>{self.pattern_year})
                {self.pattern_separator}
                (?P<month>{self.pattern_months})
            )
            ([^0-9].*|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            month = int(self._month_to_number(match.group("month")))
            try:
                return (
                    date(int(match.group("year")), month, 1),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def mmddyyyy(self) -> tuple[date, str] | None:
        """Search for a date in the format mmddyyyy.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<month>{self.pattern_month})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_inflexible})
                {self.pattern_separator}
                (?P<year>{self.pattern_year})
            )
            ([^0-9].*|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            try:
                return (
                    date(
                        int(match.group("year")), int(match.group("month")), int(match.group("day"))
                    ),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def ddmmyyyy(self) -> tuple[date, str] | None:
        """Search for a date in the format ddmmyyyy.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?P<found>
                (?P<day>{self.pattern_day_inflexible})
                {self.pattern_separator}
                (?P<month>{self.pattern_month})
                {self.pattern_separator}
                (?P<year>{self.pattern_year})
            )
            ([^0-9].*|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            try:
                return (
                    date(
                        int(match.group("year")), int(match.group("month")), int(match.group("day"))
                    ),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def mm_dd(self) -> tuple[date, str] | None:
        """Search for a date in the format mm-dd.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?:^|[^0-9])
            (?P<found>
                (?P<month>{self.pattern_month})
                {self.pattern_separator}
                (?P<day>{self.pattern_day_inflexible})
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            year = date.today().year
            try:
                return (
                    date(year, int(match.group("month")), int(match.group("day"))),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def dd_mm(self) -> tuple[date, str] | None:
        """Search for a date in the format dd-mm.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            rf"""
            (?:^|[^0-9])
            (?P<found>
                (?P<day>{self.pattern_day_inflexible})
                {self.pattern_separator}
                (?P<month>{self.pattern_month})
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            year = date.today().year
            try:
                return (
                    date(year, int(match.group("month")), int(match.group("day"))),
                    str(match.group("found")),
                )
            except ValueError as e:
                log.trace(f"Error while reformating date {match}: {e}")
                return None
        return None

    def today(self) -> tuple[date, str] | None:
        """Search for a date in the format today.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            r"""
            (?:^|[^0-9])
            (?P<found>
                (?P<today>today'?s?)
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            return (
                date.today(),
                str(match.group("found")),
            )
        return None

    def yesterday(self) -> tuple[date, str] | None:
        """Search for a date in the format yesterday.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            r"""
            (?:^|[^0-9])
            (?P<found>
                (?P<yesterday>yesterday'?s?)
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            yesterday = date.today() - timedelta(days=1)
            return (
                date(yesterday.year, yesterday.month, yesterday.day),
                str(match.group("found")),
            )
        return None

    def tomorrow(self) -> tuple[date, str] | None:
        """Search for a date in the format tomorrow.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            r"""
            (?:^|[^0-9])
            (?P<found>
                (?P<tomorrow>tomorrow'?s?)
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            tomorrow = date.today() + timedelta(days=1)
            return (
                date(tomorrow.year, tomorrow.month, tomorrow.day),
                str(match.group("found")),
            )
        return None

    def last_week(self) -> tuple[date, str] | None:
        """Search for a date in the format last week.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            r"""
            (?:^|[^0-9])
            (?P<found>
                (?P<last_week>last[- _\.]?week'?s?)
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            return (
                date.today() - timedelta(days=7),
                str(match.group("found")),
            )
        return None

    def last_month(self) -> tuple[date, str] | None:
        """Search for a date in the format last month.

        Args:
            string (str): String to search for a date.

        Returns:
            tuple[date, str]: A tuple containing the date and the date string found.
        """
        pattern = re.compile(
            r"""
            (?:^|[^0-9])
            (?P<found>
                (?P<last_month>last[- _\.]?month'?s?)
            )
            (?:[^0-9]|$)
            """,
            re.X | re.I,
        )
        match = pattern.search(self.string)
        if match:
            return (
                date.today().replace(month=date.today().month - 1, day=1),
                str(match.group("found")),
            )
        return None


class Date:
    """Date class for jdfile."""

    def __init__(self, date_format: str, string: str, ctime: datetime = None) -> None:
        """Initialize the Date class.

        Args:
            date_format (str): Date format to use.
            string (str): String to search for a date.
            ctime (datetime, optional): Creation time of the file. Defaults to None.
        """
        self.original_string = string
        self.date_format = date_format
        self.ctime = ctime
        if self.date_format is None:
            self.date, self.found_string, self.reformatted_date = None, None, None
        else:
            self.date, self.found_string = self._find_date()
            self.reformatted_date = self._reformat_date()

    def _find_date(self) -> tuple:  # noqa: C901,PLR0911,PLR0912
        """Find date in a string and reformat it to self.date_format. If no date is found, return None.

        Args:
            text (str): The text to search for dates

        Returns:
            (tuple) A tuple containing the reformatted date and the date string found in the input.
        """
        date_search = DatePattern(self.original_string)
        if date_search.yyyy_mm_dd():
            return date_search.yyyy_mm_dd()

        if date_search.yyyy_dd_mm():  # pragma: no cover
            return date_search.yyyy_dd_mm()

        if date_search.month_dd_yyyy():  # pragma: no cover
            return date_search.month_dd_yyyy()

        if date_search.dd_month_yyyy():  # pragma: no cover
            return date_search.dd_month_yyyy()

        if date_search.month_dd():  # pragma: no cover
            return date_search.month_dd()

        if date_search.month_yyyy():  # pragma: no cover
            return date_search.month_yyyy()

        if date_search.yyyy_month():  # pragma: no cover
            return date_search.yyyy_month()

        if date_search.mmddyyyy():  # pragma: no cover
            return date_search.mmddyyyy()

        if date_search.ddmmyyyy():  # pragma: no cover
            return date_search.ddmmyyyy()

        if date_search.mm_dd():  # pragma: no cover
            return date_search.mm_dd()

        if date_search.dd_mm():  # pragma: no cover
            return date_search.dd_mm()

        if date_search.today():  # pragma: no cover
            return date_search.today()

        if date_search.yesterday():  # pragma: no cover
            return date_search.yesterday()

        if date_search.tomorrow():  # pragma: no cover
            return date_search.tomorrow()

        if date_search.last_week():  # pragma: no cover
            return date_search.last_week()

        if date_search.last_month():  # pragma: no cover
            return date_search.last_month()

        if self.ctime:
            return date(self.ctime.year, self.ctime.month, self.ctime.day), None

        return None, None

    def _reformat_date(self) -> str:
        """Reformat the date to self.date_format.

        Returns:
            str: Reformatted date.
        """
        if self.date:
            try:
                return self.date.strftime(self.date_format)
            except ValueError as e:
                log.trace(f"Error while reformating date {self.date}: {e}")
                self.date, self.found_string, self.reformatted_date = None, None, None
        return None



"""Enums for the jdfile app."""

from enum import Enum


class FolderType(Enum):
    """Enum for folder types."""

    AREA = "area"
    CATEGORY = "category"
    SUBCATEGORY = "subcategory"


class TransformCase(Enum):
    """Define choices for case transformation."""

    CAMELCASE = "camelcase"
    IGNORE = "ignore"
    LOWER = "lower"
    SENTENCE = "sentence"
    TITLE = "title"
    UPPER = "upper"


class Separator(Enum):
    """Define choices for separator transformation."""

    DASH = "-"
    IGNORE = "ignore"
    NONE = ""
    SPACE = " "
    UNDERSCORE = "_"


class InsertLocation(Enum):
    """Define choices for inserting text."""

    AFTER = "after"
    BEFORE = "before"



"""Model for the File object."""
import difflib
import re
from datetime import datetime
from pathlib import Path
from typing import Any

import rich.repr

from jdfile._config import Config
from jdfile.models.dates import Date
from jdfile.models.project import Project
from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.nltk import find_synonyms
from jdfile.utils.questions import select_folder
from jdfile.utils.strings import (
    insert,
    match_case,
    normalize_separators,
    split_camelcase_words,
    split_words,
    strip_special_chars,
    strip_stopwords,
    transform_case,
)


@rich.repr.auto
class File:
    """Representation for a File object."""

    def __init__(
        self, path: Path = None, config: Config = Config(), project: Project = None
    ) -> None:
        """Initialize the File object.

        Args:
            path (Path, optional): Path to the file. Defaults to None.
            config (Config, optional): Configuration object. Defaults to Config().
            project (Project, optional): Project object. Defaults to None.
        """
        log.trace(f"Initializing File: {path}")
        self.config = config
        self.accessed = datetime.fromtimestamp(path.stat().st_atime)
        self.created = datetime.fromtimestamp(path.stat().st_ctime)
        self.modified = datetime.fromtimestamp(path.stat().st_mtime)
        self.name = path.name
        self.parent = path.parent
        self.path = path.expanduser().resolve()
        self.suffixes = self.path.suffixes
        self.stem = str(self.path)[: str(self.path).rfind("".join(self.path.suffixes))].replace(
            f"{str(self.parent)}/", ""
        )
        if re.match(r"^\.", self.stem):
            self.is_dotfile: bool = True
        else:
            self.is_dotfile = False
        self.project = project
        self.organize_possible_folders: dict[str, Any] = {}
        self.organize_skip = False

        #### Create new attributes for the file ####
        if self.config.date_format:
            self.date = Date(date_format=config.date_format, string=self.stem, ctime=self.created)
        else:
            self.date = None

        if self.config.clean:
            self.new_stem = self._clean_stem()
            self.new_suffixes = self._clean_suffixes()
        else:
            self.new_stem = self.stem
            if self.date and self.date.found_string:
                self.new_stem = re.sub(re.escape(self.date.found_string), "", self.new_stem)
            if self.date and self.date.reformatted_date:
                self.new_stem = insert(
                    self.new_stem,
                    self.date.reformatted_date,
                    self.config.insert_location,
                    self.config.separator,
                )
            self.new_suffixes = self.suffixes.copy()

        self.new_parent = (
            self._get_new_parent()
            if self.project and self.project.exists and config.organize
            else self.parent
        )

    def __rich_repr__(self) -> rich.repr.RichReprResult:  # pragma: no cover
        """Rich representation of the File object.

        Returns:
            rich.repr.RichReprResult: Rich representation of the File object.
        """
        yield "accessed", self.accessed
        yield "created", self.created
        yield "is_dotfile", self.is_dotfile
        yield "modified", self.modified
        yield "name", self.name
        yield "parent", self.parent
        yield "path", self.path
        yield "skip_file", self.organize_skip
        yield "stem", self.stem
        yield "suffixes", self.suffixes
        yield "target", self.target

    def _clean_stem(self) -> str:
        """Create a cleaned stem for the file.

        Returns:
            str: Cleaned stem for the file.
        """
        new_stem = self.stem

        if self.date and self.date.found_string:
            new_stem = re.sub(re.escape(self.date.found_string), "", new_stem)

        if self.config.split_words:
            new_stem = split_camelcase_words(new_stem, self.config.match_case)

        if self.config.strip_stopwords:
            new_stem = strip_stopwords(new_stem, self.config.stopwords)

        new_stem = strip_special_chars(new_stem)
        new_stem = transform_case(new_stem, self.config.transform_case)
        new_stem = match_case(new_stem, self.config.match_case)
        new_stem = normalize_separators(new_stem, self.config.separator)
        new_stem = new_stem.strip(" -_.")

        if self.date and self.date.reformatted_date:
            new_stem = insert(
                new_stem,
                self.date.reformatted_date,
                self.config.insert_location,
                self.config.separator,
            )

        if self.is_dotfile and not re.match(r"^\.", new_stem):
            return f".{new_stem}"

        return new_stem

    def _clean_suffixes(self) -> list[str]:
        """Clean suffixes from the new filename.

        Returns:
            list[str]: Cleaned suffixes.
        """
        return [
            ".jpg" if ext == ".jpeg" or ext == ".JPEG" else ext.lower() for ext in self.suffixes
        ]

    def _get_new_parent(self) -> Path:
        """Identify the new parent directory for the file.

        Returns:
            Path: New directory for the file.
        """
        words_in_stem = self.tokenize_stem()

        # If a JD number is specified, stop matching other terms
        if match_from_number := [
            i.path for i in self.project.usable_folders if i.number in words_in_stem
        ]:
            log.trace(f"'{self.name}' matched folder from number")
            return match_from_number[0]

        # Match terms to folders
        possible_folders: dict[str, Any] = {}
        for f in self.project.usable_folders:
            key = str(f.path.relative_to(self.project.path))
            for term in [_t.lower() for _t in f.terms]:
                if term.lower() in words_in_stem and key not in possible_folders:
                    possible_folders[key] = [f, [term]]
                if term.lower() in words_in_stem and key in possible_folders:
                    possible_folders[key][1].append(term)

        if len(possible_folders) == 0:
            self.organize_skip = True
            return self.parent

        if len(possible_folders) == 1 or self.config.force:
            log.trace(f"'{self.name}': 1 matching folder")
            return list(possible_folders.values())[0][0].path

        if len(possible_folders) > 1:
            self.organize_possible_folders = possible_folders
            return None

        # If no matches are found, skip the file
        self.organize_skip = True
        return self.parent

    @property
    def target(self) -> Path:
        """Return the target path for the file.

        Returns:
            Path: Target path for the file.
        """
        return Path(self.new_parent / f"{self.new_stem}{''.join(self.new_suffixes)}")

    def commit(self, verbosity: int = 1) -> bool:
        """Commit the changes to the file by writing the new filename to disk.

        Returns:
            bool: True if the file was successfully moved, False otherwise.
            verbosity (int, optional): Verbosity level. Defaults to 1.
        """
        if not self.has_changes() or self.organize_skip:
            if verbosity > 0:
                alerts.info(f"{self.name} -> No changes")
            return False

        if self.target.exists() and (not self.config.overwrite_existing or self.target.is_dir()):
            self.unique_name()

        if self.project and self.project.exists:
            try:
                display = ".../" + str(self.target.relative_to(self.project.path.parents[0]))
            except ValueError:  # pragma: no cover
                display = str(self.target)
        else:
            display = self.target.name

        if self.config.dry_run:
            alerts.dryrun(f"{self.name} -> {display}")
            return True

        self.path.rename(self.target)
        log.success(f"{self.name} -> {display}")
        return True

    def has_changes(self) -> bool:
        """Return whether the file has a change.

        Returns:
            True if the file has a change, False otherwise.
        """
        return self.target != self.path

    def print_diff(self) -> str:
        """Print the diff between the original and new stems.

        Returns:
            str: Diff between the original and new stems.
        """
        output = []
        a = f"{self.stem}{''.join(self.suffixes)}"
        b = f"{self.new_stem}{''.join(self.new_suffixes)}"

        matcher = difflib.SequenceMatcher(None, a, b)

        green = "[green reverse]"
        red = "[red reverse]"
        endgreen = "[/]"
        endred = "[/]"

        for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
            if opcode == "equal":
                output.append(a[a0:a1])
            elif opcode == "insert":
                output.append(f"{green}{b[b0:b1]}{endgreen}")
            elif opcode == "delete":
                output.append(f"{red}{a[a0:a1]}{endred}")
            elif opcode == "replace":
                output.append(f"{green}{b[b0:b1]}{endgreen}")
                output.append(f"{red}{a[a0:a1]}{endred}")

        return "".join(output)

    def select_new_parent(self) -> None:
        """Select the new parent directory for the file."""
        result = select_folder(
            possible_folders=self.organize_possible_folders,
            project_path=self.project.path,
            filename=self.name,
        )
        if result == "skip":
            self.organize_skip = True
            self.new_parent = self.parent
        else:
            self.organize_skip = False
            self.new_parent = Path(result)

    def tokenize_stem(self) -> list[str]:
        """Tokenize the stem of the file.

        Returns:
            list[str]: List of tokens from the stem, all lowercase.
        """
        string = self.new_stem
        string = split_camelcase_words(string)
        words_in_stem = split_words(string)
        words_in_stem.extend(self.config.cli_terms)
        synonyms = []
        if self.config.use_nltk:  # pragma: no cover
            for _w in words_in_stem:
                synonyms.extend(find_synonyms(_w))
        words_in_stem.extend(synonyms)

        return sorted({x.lower() for x in words_in_stem})

    def unique_name(self) -> None:
        """Append an integer to the end of the new filename if the target path already exists."""
        sep = self.config.separator.value if self.config.separator.name != "IGNORE" else "_"

        i = 1
        original_new_stem = f"{self.new_stem}"
        while self.target.exists():
            self.new_stem = f"{original_new_stem}{sep}{i}"
            i += 1



"""Work with NLTK Library."""

from pathlib import Path

import nltk

from jdfile.utils.alerts import logger as log


def instantiate_nltk() -> None:  # pragma: no cover
    """Instantiate nltk package."""
    ntlk_data_path = Path(Path.home() / ".jdfile" / "nltk_data")
    nltk.data.path.append(ntlk_data_path)

    install = False
    if Path(ntlk_data_path / "corpora" / "wordnet.zip").exists() is False:
        install = True
        nltk.download("wordnet", download_dir=ntlk_data_path)

    if Path(ntlk_data_path / "corpora" / "omw-1.4.zip").exists() is False:
        install = True
        nltk.download("omw-1.4", download_dir=ntlk_data_path)

    if install:
        log.success("NLTK English synonym library instantiated")
    else:
        log.trace("NLTK English synonym library already installed")


def find_synonyms(word: str) -> list[str]:  # pragma: no cover
    """Find synonyms for a word.

    Args:
        word (str): The word to find synonyms for.

    Returns:
        list[str]: De-duped alphabetical list of synonyms.
    """
    from nltk.corpus import wordnet

    synonyms = [word]

    if len(wordnet.synsets(word)) > 0:
        for syn in wordnet.synsets(word):
            for lm in syn.lemmas():
                synonyms.append(lm.name())

        for w in wordnet.synsets(word)[0].also_sees():
            synonyms.append(w.lemmas()[0].name())

        for w in wordnet.synsets(word)[0].similar_tos():
            synonyms.append(w.lemmas()[0].name())

    if word.lower() in synonyms:
        synonyms.remove(word.lower())

    return sorted(set(synonyms))



"""Model for the project directory."""
import functools
import re
from pathlib import Path
from typing import Any

import rich.repr
import typer

from jdfile._config import Config
from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.enums import FolderType


@rich.repr.auto
class Folder:
    """Representation of a folder that is available for content to be filed to.

    Attributes:
        area: (Path) Path to the area folder, if the folder is a category or subcategory.
        category: (Path) Path to the category folder, if the folder is a subcategory.
        name: (str) Name of the folder.
        number: (str) Number of the folder.
        path: (Path) Path to the folder.
        relative: (str) Path relative to the parent folder of the project
        terms: (list[str]) List of terms in the folder name and .jdfile files.
        type: (FolderType) Type of the folder.

    """

    def __init__(
        self,
        path: str,
        folder_type: FolderType,
        area: Path = None,
        category: Path = None,
    ) -> None:
        self.path = Path(path).expanduser().resolve()
        self.type = folder_type
        self.area = area
        self.category = category

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Rich representation of the Folder object."""
        yield "area", self.area
        yield "category", self.category
        yield "name", self.name
        yield "number", self.number
        yield "path", self.path
        yield "terms", self.terms
        yield "type", self.type

    @functools.cached_property
    def name(self) -> str:
        """Name of the folder."""
        if self.type == FolderType.AREA:
            return re.sub(r"^\d{2}-\d{2}[- _]", "", str(self.path.name)).strip()

        if self.type == FolderType.CATEGORY:
            return re.sub(r"^\d{2}[- _]", "", str(self.path.name)).strip()

        if self.type == FolderType.SUBCATEGORY:
            return re.sub(r"^\d{2}\.\d{2}[- _]", "", str(self.path.name)).strip()

        return None

    @functools.cached_property
    def number(self) -> str:
        """Johnny Decimal number of the folder."""
        if self.type == FolderType.AREA:
            return re.match(r"^(\d{2}-\d{2})[- _]", str(self.path.name)).group(0).strip("- _")

        if self.type == FolderType.CATEGORY:
            return re.match(r"^(\d{2})[- _]", str(self.path.name)).group(0).strip("- _")

        if self.type == FolderType.SUBCATEGORY:
            return re.match(r"^(\d{2}\.\d{2})[- _]", str(self.path.name)).group(0).strip("- _")

        return None

    @functools.cached_property
    def terms(self) -> list[str]:
        """Terms used to match the folder."""
        terms = [word for word in re.split(r"[- _]", self.name) if word]
        if Path(self.path, ".jdfile").exists():
            content = Path(self.path, ".jdfile").read_text().splitlines()
            for line in content:
                if line.startswith("#"):
                    continue
                terms.append(line)

        return terms


@rich.repr.auto
class Project:
    """Representation of a project directory.

    Attributes:
        all_folders: (dict[str, dict[str, Any]]) All folders in the project.
        config: (dict[str, Any]) Configuration for the project.
        exists: (bool) Whether the project exists.
        name: (str) Name of the project.
        path: (Path) Path to the project.
        usable_folders: (dict[str, dict[str, Any]]) Johnny Decimal folders that can be used for filing.
    """

    def __init__(
        self,
        config: Config = None,
    ) -> None:
        """Initialize the project folder."""
        self.config = config
        if (
            self.config is None
            or self.config.project_name is None
            or self.config.project_path is None
        ):
            self.path = None
            self.name = None
            self.exists = False
            self.all_folders = {}
        else:
            self.path = self._validate_project_path(self.config.project_path)
            self.name = self.config.project_name
            self.exists = self.path.exists()
            self.all_folders = self._find_folders()

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        """Rich representation of the Project object."""
        yield "exists", self.exists
        yield "name", self.name
        yield "path", self.path

    def _find_folders(self) -> dict[str, dict[str, Any]]:
        """Find all areas (top level folders) in the project.

        Returns:
            list[Path]: List of paths to areas.
        """
        area_dict = {}
        for area in self.path.iterdir():
            if area.is_dir() and re.match(r"^\d{2}-\d{2}[- _]", area.name):
                area_dict[area.name] = {
                    "path": area,
                    "categories": self._find_categories(area),
                }
        return dict(sorted(area_dict.items()))

    def _find_categories(self, area: Path) -> dict[str, dict[str, Any]]:
        """Find all categories in the project.

        Args:
            area: (Path) Path to the area.

        Returns:
            Dict[str, list[Path] | Path]: Dictionary of categories and subcategories.
        """
        category_dict = {}
        for category in area.iterdir():
            if category.is_dir() and re.match(r"^\d{2}[- _]", category.name):
                category_dict[category.name] = {
                    "path": category,
                    "subcategories": self._find_subcategories(category),
                }
        return dict(sorted(category_dict.items()))

    def _find_subcategories(self, category: Path) -> list[Path]:
        """Find all subcategories in the project.

        Args:
            category: (Path) Path to the category.

        Returns:
            list[Path]: List of paths to subcategories.
        """
        return sorted(
            [
                subcategory
                for subcategory in category.iterdir()
                if subcategory.is_dir() and re.match(r"^\d{2}\.\d{2}[- _]", subcategory.name)
            ]
        )

    def _validate_project_path(self, path: Path) -> Path:
        """Assign the project path after validating it exists.

        Args:
            path: (Path) Path to the project.

        Returns:
            bool: Whether the path is valid.
        """
        path = Path(path).expanduser().resolve()

        if not path.exists():
            alerts.error(f"Specified project path does not exist: {path}")
            raise typer.Abort()

        if not path.is_dir():
            alerts.error(f"Specified project path is not a directory: {path}")
            raise typer.Abort()

        return path

    @functools.cached_property
    def usable_folders(self) -> list[Folder]:
        """Create property for usable folders.

        Returns:
            list[Folder]: List of paths to usable folders.
        """
        if not self.exists:
            return []

        usable_folders: list[Folder] = []
        for _area in self.all_folders:
            if (
                self.all_folders[_area]["categories"] == {}
                or Path(self.all_folders[_area]["path"] / ".jdfile").exists()
            ):
                usable_folders.append(
                    Folder(
                        self.all_folders[_area]["path"],
                        FolderType.AREA,
                        area=self.all_folders[_area]["path"],
                    )
                )

            if self.all_folders[_area]["categories"] != {}:
                for _category in self.all_folders[_area]["categories"]:
                    if (
                        self.all_folders[_area]["categories"][_category]["subcategories"] == []
                        or Path(
                            self.all_folders[_area]["categories"][_category]["path"] / ".jdfile"
                        ).exists()
                    ):
                        usable_folders.append(
                            Folder(
                                self.all_folders[_area]["categories"][_category]["path"],
                                FolderType.CATEGORY,
                                area=self.all_folders[_area]["path"],
                                category=self.all_folders[_area]["categories"][_category]["path"],
                            )
                        )

                    if len(self.all_folders[_area]["categories"][_category]["subcategories"]) > 0:
                        for _subcategory in self.all_folders[_area]["categories"][_category][
                            "subcategories"
                        ]:
                            usable_folders.append(
                                Folder(
                                    _subcategory,
                                    FolderType.SUBCATEGORY,
                                    area=self.all_folders[_area]["path"],
                                    category=self.all_folders[_area]["categories"][_category][
                                        "path"
                                    ],
                                ),
                            )
        log.trace(f"Populated {len(usable_folders)} folders")
        return sorted(usable_folders, key=lambda x: x.path)

    def tree(self) -> None:
        """Print a tree of the project."""
        pipe = "│"
        branch = "├──"
        space = "   "
        elbow = "└──"

        if self.exists is False:
            alerts.warning("No Johnny Decimal project found.")
            return
        if self.exists:
            print(self.path)
            print("│")
            for _n, _area in enumerate(self.all_folders):
                if _n < len(self.all_folders) - 1:
                    print(branch, _area)
                    last_area = False
                else:
                    print(elbow, _area)
                    last_area = True
                for _nn, _category in enumerate(self.all_folders[_area]["categories"]):
                    area_pipe = space if last_area else pipe
                    if _nn < len(self.all_folders[_area]["categories"]) - 1:
                        print(area_pipe, space, branch, _category)
                        last_cat = False
                    else:
                        print(area_pipe, space, elbow, _category)
                        last_cat = True

                    for _nnn, _subcategory in enumerate(
                        self.all_folders[_area]["categories"][_category]["subcategories"]
                    ):
                        cat_pipe = space if last_cat else pipe
                        if (
                            _nnn
                            < len(self.all_folders[_area]["categories"][_category]["subcategories"])
                            - 1
                        ):
                            print(area_pipe, space, cat_pipe, space, branch, _subcategory.name)
                        else:
                            print(area_pipe, space, cat_pipe, space, elbow, _subcategory.name)



"""String utilities."""
import re
from contextlib import suppress

from jdfile.utils.alerts import logger as log
from jdfile.utils.enums import InsertLocation, Separator, TransformCase


def insert(string: str, value: str, location: InsertLocation, separator: Separator) -> str:
    """Insert a value into the new filename.

    Args:
        string (str): String to insert value into.
        location (InsertLocation): Where to insert the value.
        separator (Separator): Separator to use.
        value (str): Value to insert.

    Returns:
        str: New filename with value inserted.
    """
    sep = separator.value if separator.name != "IGNORE" else "_"

    match location:
        case InsertLocation.BEFORE:
            return f"{value}{sep}{string}"
        case InsertLocation.AFTER:
            return f"{string}{sep}{value}"


def match_case(string: str, match_case: list[str] = []) -> str:
    """Match case of words in a string to a list of words.

    Args:
        match_case (list[str]): List of words to match case.
        string (str): String to match case in.

    Returns:
        str: String with case matched.
    """
    if len(match_case) > 0:
        for term in match_case:
            with suppress(re.error):
                string = re.sub(
                    rf"(^|[-_ ]){re.escape(term)}([-_ ]|$)", rf"\1{term}\2", string, flags=re.I
                )

    return string


def normalize_separators(string: str, separator: Separator = Separator.IGNORE) -> str:
    """Normalize separators in a string.

    Args:
        separator (Separator): Separator to normalize to.
        string (str): String to fix separators in.

    Returns:
        str: String with separators normalized.
    """
    if separator.name != "IGNORE":
        return re.sub(r"[-_ \.]+", separator.value, string).strip("-_ ")

    return re.sub(r"([-_ \.])[-_ \.]+", r"\1", string).strip("-_ ")


def split_camelcase_words(string: str, match_case: list[str] = []) -> str:
    """Split camelCase words in a string into separate words.

    If camelCase words are in the match_case list, they will be put back together.

    Args:
        match_case (list[str], optional): List of words to retain case. Defaults to [].
        string (str): String to split words in.

    Returns:
        str: String with camelCase words split.
    """
    words = " ".join([word for word in re.split(r"(?=[A-Z][a-z])", string) if word])

    # Put match case words back together
    if len(match_case) > 0:
        match_case_terms = {}
        for _match in match_case:
            split_term = " ".join([w for w in re.split(r"(?=[A-Z][a-z])", _match) if w])
            match_case_terms[_match] = split_term

        for phrase, split_phrase in match_case_terms.items():
            words = re.sub(
                rf"(^|[-_ \d]){re.escape(split_phrase)}([-_ \d]|$)",
                rf"\1{phrase}\2",
                words,
                flags=re.I,
            )

    return words


def split_words(
    string: str,
) -> list[str]:
    """Split a string into a list of words. Ignore single letters, special characters, and numbers.

    Args:
        string (str): String to split into words.

    Returns:
        list[str]: List of words.
    """
    string = strip_special_chars(string)
    return [
        w for w in re.split(r"[-_ \.]+", string) if re.match(r"^\d*[A-Z]+\d*[A-Z]+\d*$", w.upper())
    ]


def strip_special_chars(string: str, replacement: str = "") -> str:
    """Strip special characters from a string.

    Args:
        string (str): String to strip special characters from.
        replacement (str, optional): Replacement character. Defaults to "".

    Returns:
        str: String with special characters stripped.
    """
    return re.sub(r"[^\w\d_ -]", replacement, string)


def strip_stopwords(string: str, stopwords: list[str] = []) -> str:
    """Strip stopwords from the new filename.

    Args:
        string(str): String to strip stopwords from.
        stopwords (list[str], optional): List of additional stopwords to strip. Defaults to [].

    Returns:
        str: String with stopwords stripped.
    """
    common_english_stopwords = [
        "a",
        "able",
        "ableabout",
        "about",
        "above",
        "abroad",
        "abst",
        "accordance",
        "according",
        "accordingly",
        "across",
        "act",
        "actually",
        "ad",
        "added",
        "adj",
        "adopted",
        "ae",
        "af",
        "affected",
        "affecting",
        "affects",
        "after",
        "afterwards",
        "ag",
        "again",
        "against",
        "ago",
        "ah",
        "ahead",
        "ai",
        "ain't",
        "aint",
        "al",
        "all",
        "allow",
        "allows",
        "almost",
        "alone",
        "along",
        "alongside",
        "already",
        "also",
        "although",
        "always",
        "am",
        "amid",
        "amidst",
        "among",
        "amongst",
        "amoungst",
        "amount",
        "an",
        "and",
        "announce",
        "another",
        "any",
        "anybody",
        "anyhow",
        "anymore",
        "anyone",
        "anything",
        "anyway",
        "anyways",
        "anywhere",
        "ao",
        "apart",
        "apparently",
        "appear",
        "appreciate",
        "appropriate",
        "approximately",
        "aq",
        "ar",
        "are",
        "area",
        "areas",
        "aren",
        "aren't",
        "arent",
        "arise",
        "around",
        "arpa",
        "as",
        "aside",
        "ask",
        "asked",
        "asking",
        "asks",
        "associated",
        "at",
        "au",
        "auth",
        "available",
        "aw",
        "away",
        "awfully",
        "az",
        "b",
        "ba",
        "back",
        "backed",
        "backing",
        "backs",
        "backward",
        "backwards",
        "bb",
        "bd",
        "be",
        "became",
        "because",
        "become",
        "becomes",
        "becoming",
        "been",
        "before",
        "beforehand",
        "began",
        "begin",
        "beginning",
        "beginnings",
        "begins",
        "behind",
        "being",
        "beings",
        "believe",
        "below",
        "beside",
        "besides",
        "better",
        "between",
        "beyond",
        "bf",
        "bg",
        "bh",
        "bi",
        "big",
        "bill",
        "billion",
        "biol",
        "bj",
        "bm",
        "bn",
        "bo",
        "both",
        "bottom",
        "br",
        "brief",
        "briefly",
        "bs",
        "bt",
        "but",
        "buy",
        "bv",
        "bw",
        "by",
        "bz",
        "c",
        "c'mon",
        "c's",
        "ca",
        "call",
        "came",
        "can",
        "can't",
        "cannot",
        "cant",
        "caption",
        "cause",
        "causes",
        "cc",
        "cd",
        "certain",
        "certainly",
        "cf",
        "cg",
        "ch",
        "ci",
        "ck",
        "cl",
        "clear",
        "clearly",
        "click",
        "cm",
        "cmon",
        "cn",
        "co",
        "co.",
        "com",
        "come",
        "comes",
        "computer",
        "con",
        "concerning",
        "consequently",
        "consider",
        "considering",
        "contain",
        "containing",
        "contains",
        "copy",
        "corresponding",
        "could",
        "could've",
        "couldn",
        "couldn't",
        "couldnt",
        "course",
        "cr",
        "cry",
        "cs",
        "cu",
        "currently",
        "cv",
        "cx",
        "cy",
        "cz",
        "d",
        "dare",
        "daren't",
        "darent",
        "de",
        "dear",
        "definitely",
        "describe",
        "described",
        "despite",
        "did",
        "didn",
        "didn't",
        "didnt",
        "differ",
        "different",
        "differently",
        "directly",
        "dj",
        "dk",
        "dm",
        "do",
        "does",
        "doesn",
        "doesn't",
        "doesnt",
        "doing",
        "don",
        "don't",
        "done",
        "dont",
        "doubtful",
        "down",
        "downed",
        "downs",
        "downwards",
        "due",
        "during",
        "dz",
        "e",
        "each",
        "early",
        "ec",
        "ed",
        "edu",
        "ee",
        "effect",
        "eg",
        "eh",
        "either",
        "eleven",
        "else",
        "elsewhere",
        "empty",
        "end",
        "ended",
        "ending",
        "ends",
        "enough",
        "entirely",
        "er",
        "es",
        "especially",
        "et",
        "et-al",
        "etc",
        "even",
        "evenly",
        "ever",
        "evermore",
        "every",
        "everybody",
        "everyone",
        "everything",
        "everywhere",
        "ex",
        "exactly",
        "example",
        "except",
        "f",
        "face",
        "faces",
        "fairly",
        "far",
        "farther",
        "felt",
        "few",
        "fewer",
        "ff",
        "fi",
        "fifth",
        "fify",
        "fill",
        "find",
        "finds",
        "fire",
        "first",
        "fix",
        "fj",
        "fk",
        "fm",
        "fo",
        "followed",
        "following",
        "follows",
        "for",
        "forever",
        "former",
        "formerly",
        "forth",
        "forty",
        "forward",
        "found",
        "fr",
        "free",
        "from",
        "front",
        "full",
        "fully",
        "further",
        "furthered",
        "furthering",
        "furthermore",
        "furthers",
        "fx",
        "g",
        "ga",
        "gave",
        "gb",
        "gd",
        "ge",
        "general",
        "generally",
        "get",
        "gets",
        "getting",
        "gf",
        "gg",
        "gh",
        "gi",
        "give",
        "given",
        "gives",
        "giving",
        "gl",
        "gm",
        "gmt",
        "gn",
        "go",
        "goes",
        "going",
        "gone",
        "good",
        "goods",
        "got",
        "gotten",
        "gov",
        "gp",
        "gq",
        "gr",
        "great",
        "greater",
        "greatest",
        "greetings",
        "grouping",
        "gs",
        "gt",
        "gu",
        "gw",
        "gy",
        "h",
        "had",
        "hadn't",
        "hadnt",
        "half",
        "happens",
        "hardly",
        "has",
        "hasn",
        "hasn't",
        "hasnt",
        "have",
        "haven",
        "haven't",
        "havent",
        "having",
        "he",
        "he'd",
        "he'll",
        "he's",
        "hed",
        "hell",
        "hello",
        "help",
        "hence",
        "her",
        "here",
        "here's",
        "hereafter",
        "hereby",
        "herein",
        "heres",
        "hereupon",
        "hers",
        "herself",
        "herse”",
        "hes",
        "hi",
        "hid",
        "highest",
        "him",
        "himself",
        "himse”",
        "his",
        "hither",
        "hk",
        "hm",
        "hn",
        "hopefully",
        "how",
        "how'd",
        "how'll",
        "how's",
        "howbeit",
        "however",
        "hr",
        "ht",
        "htm",
        "html",
        "hu",
        "i",
        "i'd",
        "i'll",
        "i'm",
        "i've",
        "i.e.",
        "id",
        "ie",
        "if",
        "ignored",
        "ii",
        "il",
        "ill",
        "im",
        "immediate",
        "immediately",
        "importance",
        "important",
        "in",
        "inasmuch",
        "inc",
        "inc.",
        "indeed",
        "indicate",
        "indicated",
        "indicates",
        "inner",
        "inside",
        "insofar",
        "instead",
        "int",
        "interest",
        "interested",
        "interesting",
        "interests",
        "into",
        "inward",
        "io",
        "iq",
        "ir",
        "is",
        "isn",
        "isn't",
        "isnt",
        "it",
        "it'd",
        "it'll",
        "it's",
        "itd",
        "itll",
        "its",
        "itself",
        "ive",
        "j",
        "je",
        "jm",
        "jo",
        "join",
        "jp",
        "just",
        "k",
        "ke",
        "keep",
        "keeps",
        "kept",
        "kg",
        "kh",
        "ki",
        "kind",
        "km",
        "kn",
        "knew",
        "know",
        "known",
        "knows",
        "kp",
        "kr",
        "kw",
        "ky",
        "kz",
        "l",
        "la",
        "largely",
        "last",
        "lately",
        "later",
        "latest",
        "latter",
        "latterly",
        "lb",
        "lc",
        "least",
        "length",
        "less",
        "lest",
        "let",
        "let's",
        "lets",
        "li",
        "like",
        "liked",
        "likely",
        "likewise",
        "line",
        "little",
        "lk",
        "ll",
        "long",
        "longer",
        "longest",
        "look",
        "looking",
        "looks",
        "lower",
        "lr",
        "ls",
        "lt",
        "ltd",
        "lu",
        "lv",
        "ly",
        "m",
        "ma",
        "made",
        "mainly",
        "make",
        "makes",
        "making",
        "man",
        "many",
        "may",
        "maybe",
        "mayn't",
        "maynt",
        "mc",
        "md",
        "me",
        "mean",
        "means",
        "meantime",
        "meanwhile",
        "member",
        "members",
        "men",
        "merely",
        "mg",
        "mh",
        "might",
        "might've",
        "mightn't",
        "mightnt",
        "mil",
        "mill",
        "million",
        "minus",
        "miss",
        "mk",
        "ml",
        "mm",
        "mn",
        "mo",
        "more",
        "moreover",
        "most",
        "mostly",
        "move",
        "mp",
        "mq",
        "mr",
        "mrs",
        "ms",
        "msie",
        "mt",
        "mu",
        "much",
        "mug",
        "must",
        "must've",
        "mustn't",
        "mustnt",
        "mv",
        "mw",
        "mx",
        "my",
        "myself",
        "mz",
        "n",
        "na",
        "namely",
        "nay",
        "nc",
        "nd",
        "ne",
        "near",
        "nearly",
        "necessarily",
        "necessary",
        "need",
        "needed",
        "needing",
        "needn't",
        "neednt",
        "needs",
        "neither",
        "never",
        "neverf",
        "neverless",
        "nevertheless",
        "newest",
        "next",
        "nf",
        "ng",
        "ni",
        "nl",
        "no",
        "no-one",
        "nobody",
        "none",
        "nonetheless",
        "noone",
        "nor",
        "normally",
        "nos",
        "noted",
        "nothing",
        "notwithstanding",
        "now",
        "nowhere",
        "np",
        "nr",
        "nu",
        "null",
        "nz",
        "o",
        "obtain",
        "obtained",
        "obviously",
        "of",
        "off",
        "often",
        "oh",
        "ok",
        "okay",
        "older",
        "oldest",
        "om",
        "omitted",
        "on",
        "once",
        "one's",
        "ones",
        "only",
        "onto",
        "open",
        "opened",
        "opening",
        "opposite",
        "or",
        "ord",
        "ordered",
        "ordering",
        "other",
        "others",
        "otherwise",
        "ought",
        "oughtn't",
        "oughtnt",
        "our",
        "ours",
        "ourselves",
        "out",
        "outside",
        "over",
        "overall",
        "owing",
        "own",
        "p",
        "pa",
        "page",
        "pages",
        "parted",
        "particular",
        "particularly",
        "parting",
        "parts",
        "past",
        "pe",
        "per",
        "perhaps",
        "pf",
        "pg",
        "ph",
        "pk",
        "pl",
        "placed",
        "please",
        "plus",
        "pm",
        "pmid",
        "pn",
        "point",
        "pointed",
        "pointing",
        "poorly",
        "possible",
        "possibly",
        "potentially",
        "pp",
        "pr",
        "predominantly",
        "present",
        "presented",
        "presenting",
        "presents",
        "presumably",
        "previously",
        "primarily",
        "probably",
        "problem",
        "problems",
        "promptly",
        "proud",
        "provided",
        "provides",
        "pt",
        "put",
        "puts",
        "pw",
        "py",
        "q",
        "qa",
        "que",
        "quickly",
        "quite",
        "qv",
        "r",
        "ran",
        "rather",
        "rd",
        "re",
        "readily",
        "really",
        "reasonably",
        "recent",
        "recently",
        "ref",
        "refs",
        "regarding",
        "regardless",
        "regards",
        "related",
        "relatively",
        "research",
        "reserved",
        "respectively",
        "resulted",
        "resulting",
        "results",
        "right",
        "ring",
        "ro",
        "round",
        "ru",
        "run",
        "rw",
        "s",
        "sa",
        "said",
        "same",
        "saw",
        "say",
        "saying",
        "says",
        "sb",
        "sc",
        "sd",
        "se",
        "sec",
        "second",
        "secondly",
        "seconds",
        "section",
        "see",
        "seeing",
        "seem",
        "seemed",
        "seeming",
        "seems",
        "seen",
        "sees",
        "self",
        "selves",
        "sensible",
        "serious",
        "seriously",
        "several",
        "sg",
        "sh",
        "shall",
        "shan't",
        "shant",
        "she",
        "she'd",
        "she'll",
        "she's",
        "shed",
        "shell",
        "shes",
        "should",
        "should've",
        "shouldn",
        "shouldn't",
        "shouldnt",
        "showed",
        "showing",
        "shown",
        "showns",
        "shows",
        "si",
        "side",
        "sides",
        "significant",
        "significantly",
        "similar",
        "similarly",
        "since",
        "sincere",
        "site",
        "sj",
        "sk",
        "sl",
        "slightly",
        "sm",
        "small",
        "smaller",
        "smallest",
        "sn",
        "so",
        "some",
        "somebody",
        "someday",
        "somehow",
        "someone",
        "somethan",
        "something",
        "sometime",
        "sometimes",
        "somewhat",
        "somewhere",
        "soon",
        "sorry",
        "specifically",
        "specified",
        "specify",
        "specifying",
        "st",
        "still",
        "stop",
        "strongly",
        "su",
        "sub",
        "substantially",
        "successfully",
        "such",
        "sufficiently",
        "suggest",
        "sup",
        "sure",
        "sv",
        "sy",
        "system",
        "sz",
        "t",
        "t's",
        "take",
        "taken",
        "taking",
        "tc",
        "td",
        "tell",
        "ten",
        "tends",
        "test",
        "text",
        "tf",
        "tg",
        "th",
        "than",
        "thank",
        "thanks",
        "thanx",
        "that",
        "that'll",
        "that's",
        "that've",
        "thatll",
        "thats",
        "thatve",
        "the",
        "their",
        "theirs",
        "them",
        "themselves",
        "then",
        "thence",
        "there",
        "there'd",
        "there'll",
        "there're",
        "there's",
        "there've",
        "thereafter",
        "thereby",
        "thered",
        "therefore",
        "therein",
        "therell",
        "thereof",
        "therere",
        "theres",
        "thereto",
        "thereupon",
        "thereve",
        "these",
        "they",
        "they'd",
        "they'll",
        "they're",
        "they've",
        "theyd",
        "theyll",
        "theyre",
        "theyve",
        "thick",
        "thin",
        "thing",
        "things",
        "think",
        "thinks",
        "third",
        "this",
        "thorough",
        "thoroughly",
        "those",
        "thou",
        "though",
        "thoughh",
        "thought",
        "thoughts",
        "thousand",
        "throug",
        "through",
        "throughout",
        "thru",
        "thus",
        "til",
        "till",
        "tip",
        "tis",
        "tj",
        "tk",
        "tm",
        "tn",
        "to",
        "today",
        "together",
        "too",
        "took",
        "top",
        "toward",
        "towards",
        "tp",
        "tr",
        "tried",
        "tries",
        "trillion",
        "truly",
        "try",
        "trying",
        "ts",
        "tt",
        "turn",
        "turned",
        "turning",
        "turns",
        "tw",
        "twas",
        "twice",
        "tz",
        "u",
        "ua",
        "ug",
        "uk",
        "um",
        "un",
        "under",
        "underneath",
        "undoing",
        "unfortunately",
        "unless",
        "unlike",
        "unlikely",
        "until",
        "unto",
        "up",
        "upon",
        "ups",
        "upwards",
        "us",
        "used",
        "useful",
        "usefully",
        "usefulness",
        "uses",
        "using",
        "usually",
        "uucp",
        "uy",
        "uz",
        "v",
        "va",
        "value",
        "various",
        "vc",
        "ve",
        "versus",
        "very",
        "vg",
        "vi",
        "via",
        "viz",
        "vn",
        "vol",
        "vols",
        "vs",
        "vu",
        "w",
        "want",
        "wanted",
        "wanting",
        "wants",
        "was",
        "wasn",
        "wasn't",
        "wasnt",
        "way",
        "ways",
        "we",
        "we'd",
        "we'll",
        "we're",
        "we've",
        "wed",
        "well",
        "went",
        "were",
        "weren",
        "weren't",
        "werent",
        "weve",
        "wf",
        "what",
        "what'd",
        "what'll",
        "what's",
        "what've",
        "whatever",
        "whatll",
        "whats",
        "whatve",
        "when",
        "when'd",
        "when'll",
        "when's",
        "whence",
        "whenever",
        "where",
        "where'd",
        "where'll",
        "where's",
        "whereafter",
        "whereas",
        "whereby",
        "wherein",
        "wheres",
        "whereupon",
        "wherever",
        "whether",
        "which",
        "whichever",
        "while",
        "whilst",
        "whim",
        "whither",
        "who",
        "who'd",
        "who'll",
        "who's",
        "whod",
        "whoever",
        "whole",
        "wholl",
        "whom",
        "whomever",
        "whos",
        "whose",
        "why",
        "why'd",
        "why'll",
        "why's",
        "widely",
        "width",
        "will",
        "willing",
        "wish",
        "with",
        "within",
        "without",
        "won",
        "won't",
        "wonder",
        "wont",
        "work",
        "worked",
        "working",
        "works",
        "world",
        "would",
        "would've",
        "wouldn",
        "wouldn't",
        "wouldnt",
        "ws",
        "www",
        "x",
        "y",
        "ye",
        "year",
        "years",
        "yes",
        "yet",
        "you",
        "you'd",
        "you'll",
        "you're",
        "you've",
        "youd",
        "youll",
        "young",
        "younger",
        "youngest",
        "your",
        "youre",
        "yours",
        "yourself",
        "yourselves",
        "youve",
        "yt",
        "yu",
        "z",
        "za",
        "zero",
        "zm",
        "zr",
    ]

    tmp_string = string
    common_english_stopwords.extend(stopwords)

    for word in common_english_stopwords:
        tmp_string = re.sub(
            rf"(^|[^A-Za-z0-9]){re.escape(word)}([^A-Za-z0-9]|$)",
            r"\1\2",
            tmp_string,
            flags=re.I,
        )

    if re.match(r"^.$|^$|^[- _]+$", tmp_string):
        log.trace(f"Skip stripping stopwords. String is empty: {string}")
        return string

    return tmp_string.strip(" -_")


def transform_case(string: str, transform_case: TransformCase) -> str:
    """Transform the case of a string.

    Args:
        string (str): String to transform case of.
        transform_case (TransformCase): Case to transform to.

    Returns:
        str: Transformed string.
    """
    match transform_case:
        case TransformCase.LOWER:
            return string.lower()
        case TransformCase.UPPER:
            return string.upper()
        case TransformCase.TITLE:
            return string.title()
        case TransformCase.CAMELCASE:
            return re.sub(r"[-_ ]", "", string.title())
        case TransformCase.SENTENCE:
            return string.capitalize()
        case _:
            return string



"""jdfile CLI."""

from pathlib import Path
from typing import Any, Optional

import typer
from rich.prompt import Confirm

from jdfile.__version__ import __version__
from jdfile._config import Config
from jdfile.models.project import Project
from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.console import console
from jdfile.utils.enums import Separator, TransformCase
from jdfile.utils.nltk import instantiate_nltk
from jdfile.utils.utilities import build_file_list, table_confirmation, table_skipped_files

app = typer.Typer(add_completion=False, no_args_is_help=True, rich_markup_mode="rich")

typer.rich_utils.STYLE_HELPTEXT = ""


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        console.print(f"{__package__} version: {__version__}")
        raise typer.Exit()


@app.command()
def main(  # noqa: C901
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        "-n",
        help="Dry run - don't actually change anything",
    ),
    clean: bool = typer.Option(
        True,
        "--clean/--no-clean",
        help="Clean filenames",
        rich_help_panel="Filename Cleaning Options",
    ),
    case: TransformCase = typer.Option(
        None,
        case_sensitive=False,
        help="Case transformation. [dim](default: ignore)[/dim]",
        rich_help_panel="Filename Cleaning Options",
        show_default=False,
    ),
    config_file: Path = typer.Option(
        Path(Path.home() / f".{__package__}/{__package__}.toml"),
        help="Specify a custom path to configuration file.",
        show_default=False,
        dir_okay=False,
        file_okay=True,
    ),
    confirm_changes: bool = typer.Option(
        False,
        "--confirm/--no-confirm",
        help="Confirm changes before committing",
        show_default=True,
    ),
    date_format: str = typer.Option(
        None,
        "--date-format",
        help="Specify a date format. If specified, the date will be appended to the filename.",
        rich_help_panel="Filename Cleaning Options",
        show_default=True,
    ),
    depth: int = typer.Option(
        1,
        "--depth",
        help="How many levels deep to search for files to process",
        rich_help_panel="Filename Cleaning Options",
        show_default=True,
    ),
    files: list[Path] = typer.Argument(
        ...,
        help="Files or directories to process",
        dir_okay=True,
        file_okay=True,
        exists=True,
        resolve_path=True,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="Force changes without prompting for confirmation. Use with caution!",
        show_default=True,
    ),
    log_file: Path = typer.Option(
        Path(Path.home() / "logs" / f"{__package__}.log"),
        help="Path to log file",
        show_default=True,
        dir_okay=False,
        file_okay=True,
        exists=False,
    ),
    log_to_file: bool = typer.Option(
        False,
        "--log-to-file",
        help="Log to file",
        show_default=True,
    ),
    organize: bool = typer.Option(
        True,
        "--organize/--no-organize",
        help="Move files into matched folders within a specified project",
        rich_help_panel="File Organization Options",
        show_default=True,
    ),
    overwrite_existing: bool = typer.Option(
        None,
        "--overwrite/--no-overwrite",
        help="Overwrite existing files",
        rich_help_panel="Filename Cleaning Options",
        show_default=True,
    ),
    print_tree: bool = typer.Option(
        False,
        "--tree",
        help="Print a tree representation of the directories within a project and exit",
        show_default=True,
        rich_help_panel="File Organization Options",
    ),
    project_name: str = typer.Option(
        None,
        "--project",
        "-p",
        help="Specify a project from the configuration file.",
        rich_help_panel="File Organization Options",
    ),
    separator: Separator = typer.Option(
        "ignore",
        "--separator",
        case_sensitive=False,
        help="Word separator",
        rich_help_panel="Filename Cleaning Options",
    ),
    split_words: bool = typer.Option(
        None,
        "--split-words/--no-split",
        help="Split words on capital letters",
        rich_help_panel="Filename Cleaning Options",
        show_default=True,
    ),
    strip_stopwords: bool = typer.Option(
        None,
        "--stopwords/--keep-stopwords",
        help="Strip stopwords when cleaning filenames.",
        rich_help_panel="Filename Cleaning Options",
        show_default=True,
    ),
    terms: list[str] = typer.Option(
        None,
        "--term",
        "-t",
        help="Term used to match files. Add multiple terms with multiple --term flags",
        rich_help_panel="File Organization Options",
    ),
    verbosity: int = typer.Option(
        0,
        "-v",
        "--verbose",
        show_default=False,
        help="""Set verbosity level (0=WARN, 1=INFO, 2=DEBUG, 3=TRACE)""",
        count=True,
    ),
    version: Optional[bool] = typer.Option(  # noqa: ARG001
        None, "--version", help="Print version and exit", callback=version_callback, is_eager=True
    ),
    use_nltk: bool = typer.Option(
        False,
        "--syns/--no-syns",
        help="Use synonyms from NLTK to help match folders. Note, this will download the NLTK corpus if it is not already installed.",
        show_default=True,
        rich_help_panel="File Organization Options",
    ),
) -> None:
    """[bold]jdfile[/] cleans and normalizes filenames. In addition, if you have directories which follow the [Johnny Decimal](https://johnnydecimal.com), jdfile can move your files into the appropriate directory.

        [bold]jdfile[/] cleans filenames based on your preferences.

    -   Remove special characters
    -   Trim multiple separators ([reverse #999999]word____word[/] becomes [reverse #999999]word_word[/])
    -   Normalize to [reverse #999999]lower case[/], [reverse #999999]upper case[/], [reverse #999999]sentence case[/], or [reverse #999999]title case[/]
    -   Normalize all files to a common word separator ([reverse #999999]_[/], [reverse #999999]-[/], [reverse #999999] [/])
    -   Enforce lowercase file extensions
    -   Remove common English stopwords
    -   Split [reverse #999999]camelCase[/] words into separate words ([reverse #999999]camel Case[/])
    -   Parse the filename for a date in many different formats
    -   Remove or reformat the date and add it to the the beginning of the filename
    -   Avoid overwriting files by adding a unique integer when renaming/moving
    -   Clean entire directory trees
    -   Optionally, show previews of changes to be made before commiting
    -   Ignore files listed in a config file by filename or by regex
    -   Specify casing for words which should never be changed (ie. [reverse #999999]iMac[/] will never be re-cased)

    When a project is specified, [bold]jdfile[/] will organize your files into folders.

    -   Move files into directory trees following the [link=https://johnnydecimal.com]Johnny Decimal[/] system
    -   Parse files and folder names looking for matching terms
    -   Uses [link=https://www.nltk.org]nltk[/] to lookup synonyms to improve matching
    -   Add [reverse #999999].jdfile[/] files to directories containing a list of words that will match files

    [bold underline]Example Usage:[/]

    [dim]Normalize all files in a directory to lowercase, with underscore separators[/]
    $ jdfile --case=lower --separator=underscore /path/to/directory

    # Clean all files in a directory and confirm all changes before committing them
    $ jdfile --clean /path/to/directory

    [dim]Strip common English stopwords from all files in a directory[/]
    $ jdfile --stopwords /path/to/directory

    [dim]Transform a date and add it to the filename[/]
    $ jdfile --date-format="%Y-%m-%d" ./somefile_march 3rd, 2022.txt

    [dim]Print a tree representation of a Johnny Decimal project[/]
    $ jdfile --project={project_name} --tree

    [dim]Use the settings of a project in the config file to clean filenames without[/]
    [dim]organizing them into folders[/]
    $ jdfile --project={project_name} --no-organize path/to/some_file.jpg

    [dim]Organize files into a Johnny Decimal project with specified terms with title casing[/]
    $ jdfile ---project={project_name} --term=term1 --term=term2 path/to/some_file.jpg
    """
    alerts.LoggerManager(  # pragma: no cover
        log_file,
        verbosity,
        log_to_file,
    )
    context: dict[str, Any] = {
        "clean": clean,
        "cli_terms": terms,
        "confirm_changes": confirm_changes,
        "date_format": date_format,
        "depth": depth,
        "dry_run": dry_run,
        "force": force,
        "organize": organize,
        "overwrite_existing": overwrite_existing,
        "print_tree": print_tree,
        "project_name": project_name,
        "separator": separator,
        "split_words": split_words,
        "strip_stopwords": strip_stopwords,
        "transform_case": case,
        "use_nltk": use_nltk,
        "verbosity": verbosity,
    }
    log.trace(f"Context: {context}")

    config: Config = Config(
        config_path=config_file,
        context=context,
    )
    log.debug(f"Loaded config: {config_file}")
    log.trace(f"Config: {config}")

    project = Project(config=config)

    log.trace(f"Loaded project: {project}")
    if print_tree:
        project.tree()
        raise typer.Exit()

    if project.exists and use_nltk:
        instantiate_nltk()
    if project.exists:
        processed_files, skip_organize_files = build_file_list(files, config, project)
    else:
        processed_files, skip_organize_files = build_file_list(files, config)

    if (
        files is None
        or len(files) == 0
        or (len(processed_files) == 0 and len(skip_organize_files) == 0)
    ):
        alerts.error("No files found to process")
        raise typer.Exit(1)

    if len(skip_organize_files) > 0:
        table_skipped_files(skip_organize_files)

    files_with_changes = [x for x in processed_files if x.has_changes()]
    if len(files_with_changes) == 0:
        alerts.info(f"No changes out of {len(processed_files)} files")
        raise typer.Exit()

    if confirm_changes and len(processed_files) > 0:
        files_to_confirm = (
            processed_files
            if context["verbosity"] > 0
            else [x for x in processed_files if x.has_changes()]
        )
        if len(files_to_confirm) == 0:
            alerts.info(f"No changes out of {len(processed_files)} files")
            raise typer.Exit()

        table_confirmation(files_to_confirm, project, total_files=len(processed_files))
        if not force:
            make_chages = Confirm.ask("Commit changes?")
            if not make_chages:
                raise typer.Exit()

    alerts.notice(f"Committing {len(files_with_changes)} of {len(processed_files)} files")
    for file in processed_files:
        file.commit(verbosity=context["verbosity"])
